#include <stdio.h>
#include <stdlib.h>
#include "mini-gmp.h"

typedef struct p2{
  mpz_t coeff[2];
}p2;

typedef struct p6{
  p2* p2coeff[3];
}p6;

typedef struct p12{
  p6* p6coeff[2];
}p12;

typedef struct ECPoint{
  mpz_t x;
  mpz_t y;
  mpz_t z;
} ECPoint;

typedef struct p2ECPoint{
  p2* x;
  p2* y;
  p2* z;
} p2ECPoint;

void p2_init(p2* p){
  mpz_init(p->coeff[0]);
  mpz_init(p->coeff[1]);
}

void p2_clear(p2* p){
  mpz_clear(p->coeff[0]);
  mpz_clear(p->coeff[1]);
}

void p2_set(p2* a, p2* b){
  mpz_set(a->coeff[0],b->coeff[0]);
  mpz_set(a->coeff[1],b->coeff[1]);
}

void p2_set_mpz_t(p2* p, mpz_t a0, mpz_t a1){
  mpz_set(p->coeff[0],a0);
  mpz_set(p->coeff[1],a1);
}

void p2_set_si(p2* p, int a0, int a1){
  mpz_t a,b;
  mpz_init(a);
  mpz_init(b);
  mpz_set_si(a,a0);
  mpz_set_si(b,a1);
  p2_set_mpz_t(p,a,b);
  mpz_clear(a);
  mpz_clear(b);
}

void p6_init(p6* p){
  p->p2coeff[0] = malloc(sizeof(p2));
  p->p2coeff[1] = malloc(sizeof(p2));
  p->p2coeff[2] = malloc(sizeof(p2));
  p2_init(p->p2coeff[0]);
  p2_init(p->p2coeff[1]);
  p2_init(p->p2coeff[2]);
}

void p6_clear(p6* p){
  p2_clear(p->p2coeff[0]);
  p2_clear(p->p2coeff[1]);
  p2_clear(p->p2coeff[2]);
  free(p->p2coeff[0]);
  free(p->p2coeff[1]);
  free(p->p2coeff[2]);
}

void p6_set(p6* a, p6* b){
  p2_set(a->p2coeff[0],b->p2coeff[0]);
  p2_set(a->p2coeff[1],b->p2coeff[1]);
  p2_set(a->p2coeff[2],b->p2coeff[2]);
}
void p6_set_p2(p6* p, p2* c0, p2* c1, p2* c2){
  p2_set(p->p2coeff[0],c0);
  p2_set(p->p2coeff[1],c1);
  p2_set(p->p2coeff[2],c2);
}

void p6_set_si(p6* p, int a0, int a1, int a2, int a3, int a4, int a5){
  mpz_t a,b;
  p2 A,B,C;
  mpz_init(a);
  mpz_init(b);
  p2_init(&A);
  p2_init(&B);
  p2_init(&C);
  mpz_set_si(a,a0);
  mpz_set_si(b,a1);
  p2_set_mpz_t(&A,a,b);
  mpz_set_si(a,a2);
  mpz_set_si(b,a3);
  p2_set_mpz_t(&B,a,b);
  mpz_set_si(a,a4);
  mpz_set_si(b,a5);
  p2_set_mpz_t(&C,a,b);
  p6_set_p2(p,&A,&B,&C);
  mpz_clear(a);
  mpz_clear(b);
  p2_clear(&A);
  p2_clear(&B);
  p2_clear(&C);
}

void p12_init(p12* p){
  p->p6coeff[0] = malloc(sizeof(p6));
  p->p6coeff[1] = malloc(sizeof(p6));
  p6_init(p->p6coeff[0]);
  p6_init(p->p6coeff[1]);
}

void p12_clear(p12* p){
  p6_clear(p->p6coeff[0]);
  p6_clear(p->p6coeff[1]);
  free(p->p6coeff[0]);
  free(p->p6coeff[1]);
}

void p12_set(p12* a, p12* b){
  p6_set(a->p6coeff[0],b->p6coeff[0]);
  p6_set(a->p6coeff[1],b->p6coeff[1]);
}

void p12_set_p6(p12* p, p6* c0, p6* c1){
  p6_set_p2(p->p6coeff[0],c0->p2coeff[0],c0->p2coeff[1],c0->p2coeff[2]);
  p6_set_p2(p->p6coeff[1],c1->p2coeff[0],c1->p2coeff[1],c1->p2coeff[2]);
}

void p12_set_si(p12* p, int a0, int a1, int a2, int a3, int a4, int a5, int a6, int a7, int a8, int a9, int a10, int a11){
  mpz_t a,b;
  p2 A,B,C;
  p6 p6A,p6B;
  mpz_init(a);
  mpz_init(b);
  p2_init(&A);
  p2_init(&B);
  p2_init(&C);
  p6_init(&p6A);
  p6_init(&p6B);
  mpz_set_si(a,a0);
  mpz_set_si(b,a1);
  p2_set_mpz_t(&A,a,b);
  mpz_set_si(a,a2);
  mpz_set_si(b,a3);
  p2_set_mpz_t(&B,a,b);
  mpz_set_si(a,a4);
  mpz_set_si(b,a5);
  p2_set_mpz_t(&C,a,b);
  p6_set_p2(&p6A,&A,&B,&C);
  mpz_set_si(a,a6);
  mpz_set_si(b,a7);
  p2_set_mpz_t(&A,a,b);
  mpz_set_si(a,a8);
  mpz_set_si(b,a9);
  p2_set_mpz_t(&B,a,b);
  mpz_set_si(a,a10);
  mpz_set_si(b,a11);
  p2_set_mpz_t(&C,a,b);
  p6_set_p2(&p6B,&A,&B,&C);
  p12_set_p6(p,&p6A,&p6B);
  mpz_clear(a);
  mpz_clear(b);
  p2_clear(&A);
  p2_clear(&B);
  p2_clear(&C);
  p6_clear(&p6A);
  p6_clear(&p6B);
}

void p2EC_init(p2ECPoint* e){
  e->x = malloc(sizeof(p2));
  e->y = malloc(sizeof(p2));
  e->z = malloc(sizeof(p2));
  p2_init(e->x);
  p2_init(e->y);
  p2_init(e->z);
}

void p2EC_clear(p2ECPoint* e){
  p2_clear(e->x);
  p2_clear(e->y);
  p2_clear(e->z);
  free(e->x);
  free(e->y);
  free(e->z);
}

void p2EC_set(p2ECPoint* P,p2ECPoint* Q){
  p2_set(P->x,Q->x);
  p2_set(P->y,Q->y);
  p2_set(P->z,Q->z);
}

void p2EC_set_p2(p2ECPoint* P,p2* a, p2* b, p2* c){
  p2_set(P->x,a);
  p2_set(P->y,b);
  p2_set(P->z,c);
}

void EC_init_0(ECPoint* e){
  mpz_init(e->x);
  mpz_init(e->y);
  mpz_init(e->z);
}

void EC_clear(ECPoint* e){
  mpz_clear(e->x);
  mpz_clear(e->y);
  mpz_clear(e->z);
}

void EC_set_mpz_t(ECPoint* e, mpz_t a, mpz_t b, mpz_t c){
  mpz_set(e->x,a);
  mpz_set(e->y,b);
  mpz_set(e->z,c);
}

void print(mpz_t X,char* str,int base){
  char* temp = malloc(sizeof(char) * 1000);
  mpz_get_str(temp,base,X);
  printf("%s%s",temp,str);
  free(temp);
}

void addmod(mpz_t res,mpz_t a, mpz_t b, mpz_t mod){
  mpz_add(res,a,b);
  mpz_mod(res,res,mod);
}

void submod(mpz_t res,mpz_t a, mpz_t b, mpz_t mod){
  mpz_sub(res,a,b);
  mpz_mod(res,res,mod);
}

void mulmod(mpz_t res,mpz_t a, mpz_t b, mpz_t mod){
  mpz_mul(res,a,b);
  mpz_mod(res,res,mod);
}

void mulmod_ui(mpz_t res,mpz_t a, unsigned int b, mpz_t mod){
  mpz_mul_ui(res,a,b);
  mpz_mod(res,res,mod);
}

void mulmod_si(mpz_t res,mpz_t a, int b, mpz_t mod){
  mpz_mul_si(res,a,b);
  mpz_mod(res,res,mod);
}

void addp2(p2* C, p2* A, p2* B,mpz_t mod){
  addmod(C->coeff[0],A->coeff[0],B->coeff[0],mod);
  addmod(C->coeff[1],A->coeff[1],B->coeff[1],mod);
}

void subp2(p2* C, p2* A, p2* B,mpz_t mod){
  submod(C->coeff[0],A->coeff[0],B->coeff[0],mod);
  submod(C->coeff[1],A->coeff[1],B->coeff[1],mod);
}

void mulp2b0(p2* C,p2* A, mpz_t b, mpz_t mod){
  mulmod(C->coeff[0],A->coeff[0],b,mod);
  mulmod(C->coeff[1],A->coeff[1],b,mod);
}

void mulp2(p2* C, p2* A, p2* B, mpz_t mod){
  mpz_t tmp,res0,res1;
  mpz_init(tmp);
  mpz_init(res0);
  mpz_init(res1);
  mulmod(tmp,A->coeff[0],B->coeff[1],mod);
  mulmod(res1,A->coeff[1],B->coeff[0],mod);
  addmod(res1,res1,tmp,mod);
  mulmod(tmp,A->coeff[1],B->coeff[1],mod);
  mulmod(res0,A->coeff[0],B->coeff[0],mod);
  submod(res0,res0,tmp,mod);
  mpz_set(C->coeff[0],res0);
  mpz_set(C->coeff[1],res1);
  mpz_clear(tmp);
  mpz_clear(res0);
  mpz_clear(res1);
}

void mulp2_ui(p2* C, p2* A, unsigned int b, mpz_t mod){
  mulmod_ui(C->coeff[0],A->coeff[0],b,mod);
  mulmod_ui(C->coeff[1],A->coeff[1],b,mod);
}

void mulp2_si(p2* C, p2* A, int b, mpz_t mod){
  mulmod_si(C->coeff[0],A->coeff[0],b,mod);
  mulmod_si(C->coeff[1],A->coeff[1],b,mod);
}

void invp2(p2* C, p2* A, mpz_t Fp,mpz_t mod){
  mpz_t t0,t1,tmp,res1;
  mpz_t beta,p_2;
  mpz_init(beta);
  mpz_init(t0);
  mpz_init(t1);
  mpz_init(tmp);
  mpz_init(res1);
  mpz_init(p_2);
  mpz_set_si(beta,-1);
  mulmod(t0,A->coeff[0],A->coeff[0],mod);
  mulmod(t1,A->coeff[1],A->coeff[1],mod);
  mulmod(tmp,beta,t1,mod);
  submod(t0,t0,tmp,mod);
  if(mpz_invert(t1,t0,mod)){
    mulmod(C->coeff[0],A->coeff[0],t1,mod);
    mulmod(res1,A->coeff[1],t1,mod);
  }else{
    mpz_sub_ui(p_2,Fp,2);
    mpz_powm(t1,t0,p_2,mod);
    mulmod(C->coeff[0],A->coeff[0],t1,mod);
    mulmod(res1,A->coeff[1],t1,mod);
  }
  mpz_mul_si(res1,res1,-1);
  mpz_set(C->coeff[1],res1);
  mpz_clear(t0);
  mpz_clear(t1);
  mpz_clear(tmp);
  mpz_clear(res1);
  mpz_clear(beta);
  mpz_clear(p_2);
}

void powp2(p2* C,p2* A, mpz_t pow,mpz_t mod){
  int size =0,i;
  p2 Copy;
  p2_init(&Copy);
  size = mpz_sizeinbase(pow,2);
  p2_set(&Copy,A);
  for (i=size-2;i>=0;i--){
    mulp2(&Copy,&Copy,&Copy,mod);
    if(mpz_tstbit(pow,i))
      mulp2(&Copy,&Copy,A,mod);
  }
  p2_set(C,&Copy);
  p2_clear(&Copy);
}

void printp2(p2* A,char* s,int base){
  print(A->coeff[0],"\t",base);
  print(A->coeff[1],s,base);
}

void printp6(p6* A,char* s,int base){
  printp2(A->p2coeff[0],"\t",base);
  printp2(A->p2coeff[1],"\t",base);
  printp2(A->p2coeff[2],s,base);
}

void addp6(p6* C, p6* A, p6* B,mpz_t mod){
  addp2(C->p2coeff[0],A->p2coeff[0],B->p2coeff[0],mod);
  addp2(C->p2coeff[1],A->p2coeff[1],B->p2coeff[1],mod);
  addp2(C->p2coeff[2],A->p2coeff[2],B->p2coeff[2],mod);
}

void subp6(p6* C, p6* A, p6* B,mpz_t mod){
  subp2(C->p2coeff[0],A->p2coeff[0],B->p2coeff[0],mod);
  subp2(C->p2coeff[1],A->p2coeff[1],B->p2coeff[1],mod);
  subp2(C->p2coeff[2],A->p2coeff[2],B->p2coeff[2],mod);
}

void mulp6(p6* C, p6* A, p6* B,mpz_t mod){
  p2 t0,t1,t2,tmp,c0,c1,c2,xi;
  p2_init(&t0);
  p2_init(&t1);
  p2_init(&t2);
  p2_init(&c0);
  p2_init(&c1);
  p2_init(&c2);
  p2_init(&tmp);
  p2_init(&xi);  
  p2_set_si(&xi,1,1);
  mulp2(&t0,A->p2coeff[0],B->p2coeff[0],mod);
  mulp2(&t1,A->p2coeff[1],B->p2coeff[1],mod);
  mulp2(&t2,A->p2coeff[2],B->p2coeff[2],mod);
  addp2(&c0,A->p2coeff[1],A->p2coeff[2],mod);
  addp2(&tmp,B->p2coeff[1],B->p2coeff[2],mod);
  mulp2(&c0,&c0,&tmp,mod);
  subp2(&c0,&c0,&t1,mod);
  subp2(&c0,&c0,&t2,mod);
  mulp2(&c0,&c0,&xi,mod);
  addp2(&c0,&c0,&t0,mod);
  addp2(&c1,A->p2coeff[0],A->p2coeff[1],mod);
  addp2(&tmp,B->p2coeff[0],B->p2coeff[1],mod);
  mulp2(&c1,&c1,&tmp,mod);
  subp2(&c1,&c1,&t0,mod);
  subp2(&c1,&c1,&t1,mod);
  mulp2(&tmp,&t2,&xi,mod);
  addp2(&c1,&c1,&tmp,mod);
  addp2(&c2,A->p2coeff[0],A->p2coeff[2],mod);
  addp2(&tmp,B->p2coeff[0],B->p2coeff[2],mod);
  mulp2(&c2,&c2,&tmp,mod);
  subp2(&c2,&c2,&t0,mod);
  subp2(&c2,&c2,&t2,mod);
  addp2(&c2,&c2,&t1,mod);
  p2_set(C->p2coeff[0],&c0);
  p2_set(C->p2coeff[1],&c1);
  p2_set(C->p2coeff[2],&c2);
  p2_clear(&xi);
  p2_clear(&c0);
  p2_clear(&c1);
  p2_clear(&c2);
  p2_clear(&t0);
  p2_clear(&t1);
  p2_clear(&t2);
  p2_clear(&tmp);
}

void mulp6_si(p6* C, p6* A, int b, mpz_t mod){
  mulp2_si(C->p2coeff[0],A->p2coeff[0],b,mod);
  mulp2_si(C->p2coeff[1],A->p2coeff[1],b,mod);
  mulp2_si(C->p2coeff[2],A->p2coeff[2],b,mod);
}

void sqp6(p6* C, p6* A, mpz_t mod){
  p2 t0,t1,t2,c0,c1,c2,xi;
  p2_init(&t0);
  p2_init(&t1);
  p2_init(&t2);
  p2_init(&c0);
  p2_init(&c1);
  p2_init(&c2);
  p2_init(&xi);  
  p2_set_si(&xi,1,1);
  mulp2(&t1,A->p2coeff[0],A->p2coeff[1],mod);
  mulp2_ui(&t1,&t1,2,mod);
  mulp2(&t2,A->p2coeff[2],A->p2coeff[2],mod);
  mulp2(&c1,&t2,&xi,mod);
  addp2(&c1,&c1,&t1,mod);
  subp2(&c2,&t1,&t2,mod);
  mulp2(&t0,A->p2coeff[0],A->p2coeff[0],mod);
  subp2(&t1,A->p2coeff[0],A->p2coeff[1],mod);
  addp2(&t1,&t1,A->p2coeff[2],mod);
  mulp2(&t2,A->p2coeff[1],A->p2coeff[2],mod);
  mulp2_ui(&t2,&t2,2,mod);
  mulp2(&t1,&t1,&t1,mod);
  mulp2(&c0,&t2,&xi,mod);
  addp2(&c0,&c0,&t0,mod);
  addp2(&c2,&c2,&t1,mod);
  addp2(&c2,&c2,&t2,mod);
  subp2(&c2,&c2,&t0,mod);
  p2_set(C->p2coeff[0],&c0);
  p2_set(C->p2coeff[1],&c1);
  p2_set(C->p2coeff[2],&c2);
  p2_clear(&xi);
  p2_clear(&t0);
  p2_clear(&t1);
  p2_clear(&t2);
  p2_clear(&c0);
  p2_clear(&c1);
  p2_clear(&c2);
}

void invp6(p6* C, p6* A, mpz_t Fp,mpz_t mod){
  p2 t0,t1,t2,t3,t4,t5,c0,c1,c2,xi;
  p2_init(&t0);
  p2_init(&t1);
  p2_init(&t2);
  p2_init(&t3);
  p2_init(&t4);
  p2_init(&t5);
  p2_init(&c0);
  p2_init(&c1);
  p2_init(&c2);
  p2_init(&xi);  
  p2_set_si(&xi,1,1);
  mulp2(&t0,A->p2coeff[1],A->p2coeff[2],mod);
  mulp2(&c0,&t0,&xi,mod);
  mulp2(&t0,A->p2coeff[0],A->p2coeff[0],mod);
  subp2(&c0,&t0,&c0,mod);
  mulp2(&t1,A->p2coeff[1],A->p2coeff[1],mod);
  mulp2(&t5,A->p2coeff[0],A->p2coeff[2],mod);
  subp2(&c2,&t1,&t5,mod);
  mulp2(&t2,A->p2coeff[2],A->p2coeff[2],mod);
  mulp2(&t4,A->p2coeff[0],A->p2coeff[1],mod);
  mulp2(&c1,&t2,&xi,mod);
  subp2(&c1,&c1,&t4,mod);
  mulp2(&t1,A->p2coeff[0],&c0,mod);
  mulp2(&t5,A->p2coeff[2],&c1,mod);
  mulp2(&t4,&t5,&xi,mod);
  addp2(&t1,&t1,&t4,mod);
  mulp2(&t5,A->p2coeff[1],&c2,mod);
  mulp2(&t4,&t5,&xi,mod);
  addp2(&t1,&t1,&t4,mod);
  invp2(&t1,&t1,Fp,mod);
  mulp2(&c0,&c0,&t1,mod);
  mulp2(&c1,&c1,&t1,mod);
  mulp2(&c2,&c2,&t1,mod);
  p2_set(C->p2coeff[0],&c0);
  p2_set(C->p2coeff[1],&c1);
  p2_set(C->p2coeff[2],&c2);
  p2_clear(&t0);
  p2_clear(&t1);
  p2_clear(&t2);
  p2_clear(&t3);
  p2_clear(&t4);
  p2_clear(&t5);
  p2_clear(&c0);
  p2_clear(&c1);
  p2_clear(&c2);
  p2_clear(&xi);
}

void printp12(p12* A,char* s,int base){
  printp6(A->p6coeff[0],"\t",base);
  printp6(A->p6coeff[1],s,base);
}

void mulp12(p12* C, p12* A, p12* B,mpz_t mod){
  p6 c0,c1,t0,t1,tmp,gamma;
  p6_init(&t0);
  p6_init(&t1);
  p6_init(&c0);
  p6_init(&c1);
  p6_init(&tmp);
  p6_init(&gamma);
  p6_set_si(&gamma,0,0,1,0,0,0);
  mulp6(&t0,A->p6coeff[0],B->p6coeff[0],mod);
  mulp6(&t1,A->p6coeff[1],B->p6coeff[1],mod);
  mulp6(&c0,&t1,&gamma,mod);
  addp6(&c0,&t0,&c0,mod);
  addp6(&c1,A->p6coeff[0],A->p6coeff[1],mod);
  addp6(&tmp,B->p6coeff[0],B->p6coeff[1],mod);
  mulp6(&c1,&c1,&tmp,mod);
  subp6(&c1,&c1,&t0,mod);
  subp6(&c1,&c1,&t1,mod);
  p6_set(C->p6coeff[0],&c0);
  p6_set(C->p6coeff[1],&c1);
  p6_clear(&t0);
  p6_clear(&t1);
  p6_clear(&c0);
  p6_clear(&c1);
  p6_clear(&tmp);
  p6_clear(&gamma);
}

void sqp12(p12* C, p12*A, mpz_t mod){
  p6 t0,t1,c0,c1,gamma;
  p6_init(&t0);
  p6_init(&t1);
  p6_init(&c0);
  p6_init(&c1);
  p6_init(&gamma);
  p6_set_si(&gamma,0,0,1,0,0,0);
  subp6(&c0,A->p6coeff[0],A->p6coeff[1],mod);
  mulp6(&t1,A->p6coeff[1],&gamma,mod);
  subp6(&t1,A->p6coeff[0],&t1,mod);
  mulp6(&t0,A->p6coeff[0],A->p6coeff[1],mod);
  mulp6(&c0,&c0,&t1,mod);
  addp6(&c0,&c0,&t0,mod);
  addp6(&c1,&t0,&t0,mod);
  mulp6(&t0,&t0,&gamma,mod);
  addp6(&c0,&c0,&t0,mod);
  p6_set(C->p6coeff[0],&c0);
  p6_set(C->p6coeff[1],&c1);
  p6_clear(&t0);
  p6_clear(&t1);
  p6_clear(&c0);
  p6_clear(&c1);
  p6_clear(&gamma);
}

void invp12(p12* C, p12*A, mpz_t Fp,mpz_t mod){
  p6 t0,t1,c0,c1;
  p2 xi,t2;
  p6_init(&t0);
  p6_init(&t1);
  p2_init(&t2);
  p6_init(&c0);
  p6_init(&c1);
  p2_init(&xi);
  p2_set_si(&xi,1,1);
  sqp6(&t0,A->p6coeff[0],mod);
  sqp6(&t1,A->p6coeff[1],mod);
  mulp2(&t2,(&t1)->p2coeff[2],&xi,mod);
  subp2((&t0)->p2coeff[0],(&t0)->p2coeff[0],&t2,mod);
  subp2((&t0)->p2coeff[1],(&t0)->p2coeff[1],(&t1)->p2coeff[0],mod);
  subp2((&t0)->p2coeff[2],(&t0)->p2coeff[2],(&t1)->p2coeff[1],mod);
  invp6(&t0,&t0,Fp,mod);
  mulp6(&c0,A->p6coeff[0],&t0,mod);
  mulp6(&c1,A->p6coeff[1],&t0,mod);
  p6_set(C->p6coeff[0],&c0);
  p6_set(C->p6coeff[1],&c1);
  p6_clear(&t0);
  p6_clear(&t1);
  p2_clear(&t2);
  p6_clear(&c0);
  p6_clear(&c1);
  p2_clear(&xi);
}

void Dblline(p2ECPoint* T, p12* l, p2ECPoint* Q, ECPoint* P,mpz_t Fp,mpz_t mod){
  mpz_t temp;
  p2 XT,YT,ZT,t0,t1,t2,t3,t4,t5,t6,tmp,T0,T1,T2;
  p2 l00,l11,l02;
  p6 l0,l1;
  mpz_init(temp);
  p6_init(&l0);
  p6_init(&l1);
  p2_init(&l11);
  p2_init(&l00);
  p2_init(&l02);
  p2_init(&t0);
  p2_init(&t1);
  p2_init(&t2);
  p2_init(&t3);
  p2_init(&t4);
  p2_init(&t5);
  p2_init(&t6);
  p2_init(&XT);
  p2_init(&YT);
  p2_init(&ZT);
  p2_init(&tmp);
  p2_init(&T0);
  p2_init(&T1);
  p2_init(&T2);
  mulp2(&t0,Q->z,Q->z,mod);
  mulp2(&t4,Q->x,Q->y,mod);
  mulp2(&t1,Q->y,Q->y,mod);
  addp2(&t3,&t0,&t0,mod);
  mpz_add_ui(temp,Fp,1);
  mpz_cdiv_q_ui(temp,temp,2);
  mulp2b0(&t4,&t4,temp,mod);
  addp2(&t5,&t0,&t1,mod);
  addp2(&t0,&t0,&t3,mod);
  addmod((&t2)->coeff[0],(&t0)->coeff[0],(&t0)->coeff[1],mod);
  submod((&t2)->coeff[1],(&t0)->coeff[1],(&t0)->coeff[0],mod);
  mulp2(&t0,Q->x,Q->x,mod);
  addp2(&t3,&t2,&t2,mod);
  addp2(&t3,&t3,&t2,mod);
  addp2(&l02,&t0,&t0,mod);
  subp2(&XT,&t1,&t3,mod);
  addp2(&l02,&l02,&t0,mod);
  addp2(&t3,&t3,&t1,mod);
  mulp2(&XT,&XT,&t4,mod);
  mulp2b0(&t3,&t3,temp,mod);
  mulp2(&T0,&t3,&t3,mod);
  mulp2(&T1,&t2,&t2,mod);
  addp2(&T2,&T1,&T1,mod);
  addp2(&t3,Q->y,Q->z,mod);
  addp2(&T2,&T1,&T2,mod);
  mulp2(&t3,&t3,&t3,mod);
  subp2(&t3,&t3,&t5,mod);
  subp2(&T0,&T0,&T2,mod);
  p2_set(&YT,&T0);
  mulp2(&ZT,&t1,&t3,mod);
  subp2(&t2,&t2,&t1,mod);
  submod((&l00)->coeff[0],(&t2)->coeff[0],(&t2)->coeff[1],mod);
  addmod((&l00)->coeff[1],(&t2)->coeff[0],(&t2)->coeff[1],mod);
  mulp2b0(&l02,&l02,P->x,mod);
  mulp2b0(&l11,&t3,P->y,mod);
  mulp2_si(&l11,&l11,-1,mod);
  p2_set((&l0)->p2coeff[0],&l00);
  p2_set((&l0)->p2coeff[2],&l02);
  p2_set((&l1)->p2coeff[1],&l11);
  p2EC_set_p2(T,&XT,&YT,&ZT);
  p12_set_p6(l,&l0,&l1);
  p6_clear(&l0);
  p6_clear(&l1);
  p2_clear(&l11);
  p2_clear(&l00);
  p2_clear(&l02);
  p2_clear(&t0);
  p2_clear(&t1);
  p2_clear(&t2);
  p2_clear(&t3);
  p2_clear(&t4);
  p2_clear(&t5);
  p2_clear(&t6);
  p2_clear(&XT);
  p2_clear(&YT);
  p2_clear(&ZT);
  p2_clear(&tmp);
  p2_clear(&T0);
  p2_clear(&T1);
  p2_clear(&T2);
  mpz_clear(temp);
}

void Addline(p2ECPoint* T, p12* l, p2ECPoint* Q, p2ECPoint* R, ECPoint* P,mpz_t mod){
  p2 XT,YT,ZT,t1,t2,t3,t4,T1,T2,l00,l02,l11,xi,sig,lamb;
  p6 l0,l1;
  p2_init(&xi);
  p6_init(&l0);
  p6_init(&l1);
  p2_init(&t1);
  p2_init(&t2);
  p2_init(&t3);
  p2_init(&t4);
  p2_init(&T1);
  p2_init(&T2);
  p2_init(&l00);
  p2_init(&l02);
  p2_init(&l11);
  p2_init(&XT);
  p2_init(&YT);
  p2_init(&ZT);
  p2_init(&sig);
  p2_init(&lamb);
  p2_set_si(&xi,1,1);
  mulp2(&t1,Q->z,R->x,mod);
  mulp2(&t2,Q->z,R->y,mod);
  subp2(&t1,Q->x,&t1,mod);
  subp2(&t2,Q->y,&t2,mod);
  mulp2(&t3,&t1,&t1,mod);
  mulp2(&XT,&t3,Q->x,mod);
  mulp2(&t4,&t2,&t2,mod);
  mulp2(&t3,&t1,&t3,mod);
  mulp2(&t4,&t4,Q->z,mod);
  addp2(&t4,&t3,&t4,mod);
  subp2(&t4,&t4,&XT,mod);
  subp2(&t4,&t4,&XT,mod);
  subp2(&XT,&XT,&t4,mod);
  mulp2(&T1,&t2,&XT,mod);
  mulp2(&T2,&t3,Q->y,mod);
  subp2(&T2,&T1,&T2,mod);
  p2_set(&YT,&T2);
  mulp2(&XT,&t1,&t4,mod);
  mulp2(&ZT,&t3,Q->z,mod);
  mulp2b0(&l02,&t2,P->x,mod);
  mulp2_si(&l02,&l02,-1,mod);
  mulp2(&T1,&t2,R->x,mod);
  mulp2(&T2,&t1,R->y,mod);
  subp2(&T1,&T1,&T2,mod);
  p2_set(&t2,&T1);
  mulp2(&l00,&t2,&xi,mod);
  mulp2b0(&l11,&t1,P->y,mod);  
  p2_set((&l0)->p2coeff[0],&l00);
  p2_set((&l0)->p2coeff[2],&l02);
  p2_set((&l1)->p2coeff[1],&l11);
  p2EC_set_p2(T,&XT,&YT,&ZT);
  p12_set_p6(l,&l0,&l1);
  p2_clear(&xi);
  p6_clear(&l0);
  p6_clear(&l1);
  p2_clear(&t1);
  p2_clear(&t2);
  p2_clear(&t3);
  p2_clear(&t4);
  p2_clear(&T1);
  p2_clear(&T2);
  p2_clear(&l00);
  p2_clear(&l02);
  p2_clear(&l11);
  p2_clear(&XT);
  p2_clear(&YT);
  p2_clear(&ZT);
  p2_clear(&sig);
  p2_clear(&lamb);
}


void powp12(p12* C,p12* A, mpz_t pow,mpz_t mod){
  int size =0,i;
  p12 Copy;
  p12_init(&Copy);
  size = mpz_sizeinbase(pow,2);
  p12_set(&Copy,A);
  for (i=size-2;i>=0;i--){
    mulp12(&Copy,&Copy,&Copy,mod);
    if(mpz_tstbit(pow,i))
      mulp12(&Copy,&Copy,A,mod);
  }
  p12_set(C,&Copy);
  p12_clear(&Copy);
}

void make_gammar(p2 gammar[5],p2 gammar2[5],p2 gammar3[5],mpz_t Fp,mpz_t mod){
  mpz_t pow;
  p2 xi;
  int i;
  p2_init(&xi);
  mpz_init(pow);
  p2_set_si(&xi,1,1);
  mpz_sub_ui(pow,Fp,1);
  mpz_divexact_ui(pow,pow,6);
  powp2(&(gammar[0]),&xi,pow,mod);
  for(i=1;i<5;i++)
    mulp2(&(gammar[i]),&(gammar[i-1]),&(gammar[0]),mod);
  for(i=0;i<5;i++){
    p2_set(&(gammar2[i]),&(gammar[i]));
    mulmod_si((&(gammar2[i]))->coeff[1],(&(gammar2[i]))->coeff[1],-1,mod);
    mulp2(&(gammar2[i]),&(gammar2[i]),&(gammar[i]),mod);
  }
  for(i=0;i<5;i++)
    mulp2(&(gammar3[i]),&(gammar[i]),&(gammar2[i]),mod);
  mpz_clear(pow);
  p2_clear(&xi);
}

void Frob(p12* C, p12* A, p2 gammar[5], mpz_t mod){
  p6 c0,c1,a0,a1;
  p2 c00,c01,c02,c10,c11,c12,a00,a01,a02,a10,a11,a12;
  p6_init(&c0);
  p6_init(&c1);
  p6_init(&a0);
  p6_init(&a1);
  p2_init(&c00);
  p2_init(&c01);
  p2_init(&c02);
  p2_init(&c10);
  p2_init(&c11);
  p2_init(&c12);
  p2_init(&a00);
  p2_init(&a01);
  p2_init(&a02);
  p2_init(&a10);
  p2_init(&a11);
  p2_init(&a12);
  p6_set(&c0,C->p6coeff[0]);
  p6_set(&c1,C->p6coeff[1]);
  p6_set(&a0,A->p6coeff[0]);
  p6_set(&a1,A->p6coeff[1]);
  p2_set(&c00,(&c0)->p2coeff[0]);
  p2_set(&c01,(&c0)->p2coeff[1]);
  p2_set(&c02,(&c0)->p2coeff[2]);
  p2_set(&c10,(&c1)->p2coeff[0]);
  p2_set(&c11,(&c1)->p2coeff[1]);
  p2_set(&c12,(&c1)->p2coeff[2]);
  p2_set(&a00,(&a0)->p2coeff[0]);
  p2_set(&a01,(&a0)->p2coeff[1]);
  p2_set(&a02,(&a0)->p2coeff[2]);
  p2_set(&a10,(&a1)->p2coeff[0]);
  p2_set(&a11,(&a1)->p2coeff[1]);
  p2_set(&a12,(&a1)->p2coeff[2]);
  mpz_set((&c00)->coeff[0],(&a00)->coeff[0]);
  mulmod_si((&c00)->coeff[1],(&a00)->coeff[1],-1,mod);
  mpz_set((&c01)->coeff[0],(&a01)->coeff[0]);
  mulmod_si((&c01)->coeff[1],(&a01)->coeff[1],-1,mod);
  mulp2(&c01,&c01,&(gammar[1]),mod);
  mpz_set((&c02)->coeff[0],(&a02)->coeff[0]);
  mulmod_si((&c02)->coeff[1],(&a02)->coeff[1],-1,mod);
  mulp2b0(&c02,&c02,(&(gammar[3]))->coeff[0],mod);
  mpz_set((&c10)->coeff[0],(&a10)->coeff[0]);
  mulmod_si((&c10)->coeff[1],(&a10)->coeff[1],-1,mod);
  mulp2(&c10,&c10,&(gammar[0]),mod);
  mpz_set((&c11)->coeff[0],(&a11)->coeff[0]);
  mulmod_si((&c11)->coeff[1],(&a11)->coeff[1],-1,mod);
  mulp2(&c11,&c11,&(gammar[2]),mod);
  mpz_set((&c12)->coeff[0],(&a12)->coeff[0]);
  mulmod_si((&c12)->coeff[1],(&a12)->coeff[1],-1,mod);
  mulp2(&c12,&c12,&(gammar[4]),mod);
  p6_set_p2(&c0,&c00,&c01,&c02);
  p6_set_p2(&c1,&c10,&c11,&c12);
  p12_set_p6(C,&c0,&c1);
  p2_clear(&c00);
  p2_clear(&c01);
  p2_clear(&c02);
  p2_clear(&c10);
  p2_clear(&c11);
  p2_clear(&c12);
  p2_clear(&a00);
  p2_clear(&a01);
  p2_clear(&a02);
  p2_clear(&a10);
  p2_clear(&a11);
  p2_clear(&a12);
  p6_clear(&c0);
  p6_clear(&c1);
  p6_clear(&a0);
  p6_clear(&a1); 
}

void Frob2(p12* C, p12* A, p2 gammar2[5], mpz_t mod){
  p6 c0,c1,a0,a1;
  p2 c00,c01,c02,c10,c11,c12,a00,a01,a02,a10,a11,a12;
  p6_init(&c0);
  p6_init(&c1);
  p6_init(&a0);
  p6_init(&a1);
  p2_init(&c00);
  p2_init(&c01);
  p2_init(&c02);
  p2_init(&c10);
  p2_init(&c11);
  p2_init(&c12);
  p2_init(&a00);
  p2_init(&a01);
  p2_init(&a02);
  p2_init(&a10);
  p2_init(&a11);
  p2_init(&a12);
  p6_set(&c0,C->p6coeff[0]);
  p6_set(&c1,C->p6coeff[1]);
  p6_set(&a0,A->p6coeff[0]);
  p6_set(&a1,A->p6coeff[1]);
  p2_set(&c00,(&c0)->p2coeff[0]);
  p2_set(&c01,(&c0)->p2coeff[1]);
  p2_set(&c02,(&c0)->p2coeff[2]);
  p2_set(&c10,(&c1)->p2coeff[0]);
  p2_set(&c11,(&c1)->p2coeff[1]);
  p2_set(&c12,(&c1)->p2coeff[2]);
  p2_set(&a00,(&a0)->p2coeff[0]);
  p2_set(&a01,(&a0)->p2coeff[1]);
  p2_set(&a02,(&a0)->p2coeff[2]);
  p2_set(&a10,(&a1)->p2coeff[0]);
  p2_set(&a11,(&a1)->p2coeff[1]);
  p2_set(&a12,(&a1)->p2coeff[2]);
  p2_set(&c00,&a00);
  mulp2(&c01,&a01,&(gammar2[1]),mod);
  mulp2(&c02,&a02,&(gammar2[3]),mod);
  mulp2(&c10,&a10,&(gammar2[0]),mod);
  mulp2(&c11,&a11,&(gammar2[2]),mod);
  mulp2(&c12,&a12,&(gammar2[4]),mod);
  p6_set_p2(&c0,&c00,&c01,&c02);
  p6_set_p2(&c1,&c10,&c11,&c12);
  p12_set_p6(C,&c0,&c1);
  p2_clear(&c00);
  p2_clear(&c01);
  p2_clear(&c02);
  p2_clear(&c10);
  p2_clear(&c11);
  p2_clear(&c12);
  p2_clear(&a00);
  p2_clear(&a01);
  p2_clear(&a02);
  p2_clear(&a10);
  p2_clear(&a11);
  p2_clear(&a12);
  p6_clear(&c0);
  p6_clear(&c1);
  p6_clear(&a0);
  p6_clear(&a1); 
}

void Frob3(p12* C, p12* A, p2 gammar3[5], mpz_t mod){
  p6 c0,c1,a0,a1;
  p2 c00,c01,c02,c10,c11,c12,a00,a01,a02,a10,a11,a12;
  mpz_t t;
  mpz_init(t);
  p6_init(&c0);
  p6_init(&c1);
  p6_init(&a0);
  p6_init(&a1);
  p2_init(&c00);
  p2_init(&c01);
  p2_init(&c02);
  p2_init(&c10);
  p2_init(&c11);
  p2_init(&c12);
  p2_init(&a00);
  p2_init(&a01);
  p2_init(&a02);
  p2_init(&a10);
  p2_init(&a11);
  p2_init(&a12);
  p6_set(&c0,C->p6coeff[0]);
  p6_set(&c1,C->p6coeff[1]);
  p6_set(&a0,A->p6coeff[0]);
  p6_set(&a1,A->p6coeff[1]);
  p2_set(&c00,(&c0)->p2coeff[0]);
  p2_set(&c01,(&c0)->p2coeff[1]);
  p2_set(&c02,(&c0)->p2coeff[2]);
  p2_set(&c10,(&c1)->p2coeff[0]);
  p2_set(&c11,(&c1)->p2coeff[1]);
  p2_set(&c12,(&c1)->p2coeff[2]);
  p2_set(&a00,(&a0)->p2coeff[0]);
  p2_set(&a01,(&a0)->p2coeff[1]);
  p2_set(&a02,(&a0)->p2coeff[2]);
  p2_set(&a10,(&a1)->p2coeff[0]);
  p2_set(&a11,(&a1)->p2coeff[1]);
  p2_set(&a12,(&a1)->p2coeff[2]);
  mpz_set((&c00)->coeff[0],(&a00)->coeff[0]);
  mulmod_si((&c00)->coeff[1],(&a00)->coeff[1],-1,mod);
  mpz_set((&c01)->coeff[0],(&a01)->coeff[0]);
  mulmod_si((&c01)->coeff[1],(&a01)->coeff[1],-1,mod);
  mpz_set(t,(&c01)->coeff[1]);
  mpz_set((&c01)->coeff[1],(&c01)->coeff[0]);
  mulmod_si((&c01)->coeff[0],t,-1,mod); 
  mpz_set((&c02)->coeff[0],(&a02)->coeff[0]);
  mulmod_si((&c02)->coeff[1],(&a02)->coeff[1],-1,mod);
  mulp2(&c02,&c02,&(gammar3[3]),mod);
  mpz_set((&c10)->coeff[0],(&a10)->coeff[0]);
  mulmod_si((&c10)->coeff[1],(&a10)->coeff[1],-1,mod);
  mulp2(&c10,&c10,&(gammar3[0]),mod);
  mpz_set((&c11)->coeff[0],(&a11)->coeff[0]);
  mulmod_si((&c11)->coeff[1],(&a11)->coeff[1],-1,mod);
  mulp2(&c11,&c11,&(gammar3[2]),mod);
  mpz_set((&c12)->coeff[0],(&a12)->coeff[0]);
  mulmod_si((&c12)->coeff[1],(&a12)->coeff[1],-1,mod);
  mulp2(&c12,&c12,&(gammar3[4]),mod);
  p6_set_p2(&c0,&c00,&c01,&c02);
  p6_set_p2(&c1,&c10,&c11,&c12);
  p12_set_p6(C,&c0,&c1);
  p2_clear(&c00);
  p2_clear(&c01);
  p2_clear(&c02);
  p2_clear(&c10);
  p2_clear(&c11);
  p2_clear(&c12);
  p2_clear(&a00);
  p2_clear(&a01);
  p2_clear(&a02);
  p2_clear(&a10);
  p2_clear(&a11);
  p2_clear(&a12);
  p6_clear(&c0);
  p6_clear(&c1);
  p6_clear(&a0);
  p6_clear(&a1);
  mpz_clear(t);
}

void maptoCyclo(p12* C, p12* A, p2 gammar2[5],mpz_t Fp,mpz_t mod){
  p12 copy,tmp;
  p12_init(&copy);
  p12_init(&tmp);
  p6_set((&copy)->p6coeff[0],A->p6coeff[0]);
  mulp6_si((&copy)->p6coeff[1],A->p6coeff[1],-1,mod);
  invp12(&tmp,&copy,Fp,mod);
  mulp12(&copy,&copy,&tmp,mod);
  Frob2(&tmp,&copy,gammar2,mod);
  mulp12(&copy,&copy,&tmp,mod);
  p12_set(C,&copy);
  p12_clear(&copy);
  p12_clear(&tmp);
}

void final_exp(p12* fexp, p12* z,mpz_t Fp,mpz_t mod){
  p12 f,f2z,f6z,f6z2,f12z3,final;
  p12 a,b;
  p2 gammar[5],gammar2[5],gammar3[5];
  int i;
  mpz_t pow;
  mpz_init(pow);
  p12_init(&f);
  p12_init(&f2z);
  p12_init(&f6z);
  p12_init(&f6z2);
  p12_init(&f12z3);
  p12_init(&a);
  p12_init(&b);
  p12_init(&final);
  for(i=0;i<5;i++){
    p2_init(&(gammar[i]));
    p2_init(&(gammar2[i]));
    p2_init(&(gammar3[i]));
  }
  make_gammar(gammar,gammar2,gammar3,Fp,mod);
  maptoCyclo(&f,z,gammar2,Fp,mod);
  mpz_set_str(pow,"4080000000000001",16);
  powp12(&f2z,&f,pow,mod);
  sqp12(&f2z,&f2z,mod);
  sqp12(&f6z,&f2z,mod);
  mulp12(&f6z,&f6z,&f2z,mod);
  powp12(&f6z2,&f6z,pow,mod);
  sqp12(&a,&f6z2,mod);
  powp12(&f12z3,&a,pow,mod);
  mulp6_si((&f6z)->p6coeff[1],(&f6z)->p6coeff[1],-1,mod);
  mulp6_si((&f12z3)->p6coeff[1],(&f12z3)->p6coeff[1],-1,mod);
  mulp12(&a,&f12z3,&f6z2,mod);
  mulp12(&a,&a,&f6z,mod);
  mulp12(&b,&a,&f2z,mod);
  mulp12(&final,&a,&f6z2,mod);
  mulp12(&final,&final,&f,mod);
  Frob(&f2z,&b,gammar,mod);
  mulp12(&final,&final,&f2z,mod);
  Frob2(&f2z,&a,gammar2,mod);
  mulp12(&final,&final,&f2z,mod);
  mulp6_si((&f)->p6coeff[1],(&f)->p6coeff[1],-1,mod);
  mulp12(&b,&b,&f,mod);
  Frob3(&f2z,&b,gammar3,mod);
  mulp12(&final,&final,&f2z,mod);
  p12_set(fexp,&final);
  for(i=0;i<5;i++){
    p2_clear(&(gammar[i]));
    p2_clear(&(gammar2[i]));
    p2_clear(&(gammar3[i]));
  }
  p12_clear(&f);
  p12_clear(&f2z);
  p12_clear(&f6z);
  p12_clear(&f6z2);
  p12_clear(&f12z3);
  p12_clear(&a);
  p12_clear(&b);
  p12_clear(&final);
  mpz_clear(pow);
}

void OptAtePairingBN(p12* aopt, ECPoint* P, p2ECPoint* Q, mpz_t Fp,mpz_t mod){ 
  mpz_t s,pow;
  p2ECPoint T,Q1,Q2,Qcopy;
  p2 t0,t1,t2,xi,Wp,Z;
  p12 f,l;
  int i,size;
  mpz_init(s);
  mpz_init(pow);
  p2EC_init(&T);
  p2EC_init(&Qcopy);
  p2EC_init(&Q1);
  p2EC_init(&Q2);
  p12_init(&f);
  p12_init(&l);
  p2_init(&t0);
  p2_init(&Wp);
  p2_init(&Z);
  p2_init(&t1);
  p2_init(&t2);
  p2_init(&xi);
  p2_set_si(&xi,1,1);
  mpz_set_str(s,"27886288892678111236",10);
  size = mpz_sizeinbase(s,2);
  p2EC_set(&T,Q);
  p12_set_si(&f,1,0,0,0,0,0,0,0,0,0,0,0);
  p2EC_set(&Qcopy,Q);
  if(mpz_cmp_ui(s,0)<0){
    mulp2_si((&Qcopy)->y,(&Qcopy)->y,-1,mod);
    mpz_abs(s,s);
  }
  p2EC_set(&T,&Qcopy);
  for(i=size-2;i>=0;i--){
    Dblline(&T,&l,&T,P,Fp,mod);
    sqp12(&f,&f,mod);
    mulp12(&f,&f,&l,mod);
    if(mpz_tstbit(s,i)){
      Addline(&T,&l,&T,&Qcopy,P,mod);
      mulp12(&f,&f,&l,mod);
    }
  }
  p2_set(&t0,Q->x);
  p2_set(&t1,Q->y);
  p2_set(&t2,Q->z);
  mulmod_si((&t0)->coeff[1],(&t0)->coeff[1],-1,mod);
  mulmod_si((&t1)->coeff[1],(&t1)->coeff[1],-1,mod);
  mpz_sub_ui(pow,Fp,1);
  mpz_divexact_ui(pow,pow,3);
  powp2(&Wp,&xi,pow,mod);
  mulp2(&t0,&t0,&Wp,mod);
  mpz_sub_ui(pow,Fp,1);
  mpz_divexact_ui(pow,pow,2);
  powp2(&Wp,&xi,pow,mod);
  mulp2(&t1,&t1,&Wp,mod);
  p2EC_set_p2(&Q1,&t0,&t1,&t2);
  mpz_mul(pow,Fp,Fp);
  mpz_sub_ui(pow,pow,1);
  mpz_divexact_ui(pow,pow,6);
  powp2(&Z,&xi,pow,mod);
  mulp2_si(&Z,&Z,-1,mod);
  mulp2(&Z,&Z,&Z,mod);
  p2_set(&t0,Q->x);
  p2_set(&t1,Q->y);
  p2_set(&t2,Q->z);
  mulp2(&t0,&t0,&Z,mod);
  p2EC_set_p2(&Q2,&t0,&t1,&t2);
  mulp6_si((&f)->p6coeff[1],(&f)->p6coeff[1],-1,mod);
  mulp2_si((&T)->y,(&T)->y,-1,mod);
  Addline(&T,&l,&T,&Q1,P,mod);
  mulp12(&f,&f,&l,mod);
  Addline(&T,&l,&T,&Q2,P,mod);
  mulp12(&f,&f,&l,mod);
  final_exp(aopt,&f,Fp,mod);
  p2_clear(&xi);
  p2_clear(&t0);
  p2_clear(&t1);
  p2_clear(&Wp);
  p2_clear(&Z);
  p2_clear(&t2);
  p2EC_clear(&T);
  p2EC_clear(&Qcopy);
  p2EC_clear(&Q1);
  p2EC_clear(&Q2);
  p12_clear(&f);
  p12_clear(&l);
  mpz_clear(s);
  mpz_clear(pow);
}

void p12in(p12* new_f,p12* f, mpz_t mod){
  p6 new_f0,new_f1,f0,f1;
  p2 new_f00,new_f01,new_f02,new_f10,new_f11,new_f12;
  p2 f00,f01,f02,f10,f11,f12;
  p2_init(&f00);
  p2_init(&f01);
  p2_init(&f02);
  p2_init(&f10);
  p2_init(&f11);
  p2_init(&f12);
  p2_init(&new_f00);
  p2_init(&new_f01);
  p2_init(&new_f02);
  p2_init(&new_f10);
  p2_init(&new_f11);
  p2_init(&new_f12);
  p6_init(&f0);
  p6_init(&f1);
  p6_init(&new_f0);
  p6_init(&new_f1);
  p6_set(&f0,f->p6coeff[0]);
  p6_set(&f1,f->p6coeff[1]);
  p2_set(&f00,(&f0)->p2coeff[0]);
  p2_set(&f01,(&f0)->p2coeff[1]);
  p2_set(&f02,(&f0)->p2coeff[2]);
  p2_set(&f10,(&f1)->p2coeff[0]);
  p2_set(&f11,(&f1)->p2coeff[1]);
  p2_set(&f12,(&f1)->p2coeff[2]);
  mpz_mod((&new_f00)->coeff[0],(&f00)->coeff[0],mod);
  mpz_mod((&new_f00)->coeff[1],(&f00)->coeff[1],mod);
  mpz_mod((&new_f01)->coeff[0],(&f01)->coeff[0],mod);
  mpz_mod((&new_f01)->coeff[1],(&f01)->coeff[1],mod);
  mpz_mod((&new_f02)->coeff[0],(&f02)->coeff[0],mod);
  mpz_mod((&new_f02)->coeff[1],(&f02)->coeff[1],mod);
  mpz_mod((&new_f10)->coeff[0],(&f10)->coeff[0],mod);
  mpz_mod((&new_f10)->coeff[1],(&f10)->coeff[1],mod);
  mpz_mod((&new_f11)->coeff[0],(&f11)->coeff[0],mod);
  mpz_mod((&new_f11)->coeff[1],(&f11)->coeff[1],mod);
  mpz_mod((&new_f12)->coeff[0],(&f12)->coeff[0],mod);
  mpz_mod((&new_f12)->coeff[1],(&f12)->coeff[1],mod);
  p6_set_p2(&new_f0,&new_f00,&new_f01,&new_f02);
  p6_set_p2(&new_f1,&new_f10,&new_f11,&new_f12);
  p12_set_p6(new_f,&new_f0,&new_f1);
  p2_clear(&f00);
  p2_clear(&f01);
  p2_clear(&f02);
  p2_clear(&f10);
  p2_clear(&f11);
  p2_clear(&f12);
  p2_clear(&new_f00);
  p2_clear(&new_f01);
  p2_clear(&new_f02);
  p2_clear(&new_f10);
  p2_clear(&new_f11);
  p2_clear(&new_f12);
  p6_clear(&f0);
  p6_clear(&f1);
  p6_clear(&new_f0);
  p6_clear(&new_f1);
}

int p12_is_eq(p12* g,p12* f){
  p6 g0,g1,f0,f1;
  p2 g00,g01,g02,g10,g11,g12;
  p2 f00,f01,f02,f10,f11,f12;
  int res = 0;
  p2_init(&f00);
  p2_init(&f01);
  p2_init(&f02);
  p2_init(&f10);
  p2_init(&f11);
  p2_init(&f12);
  p2_init(&g00);
  p2_init(&g01);
  p2_init(&g02);
  p2_init(&g10);
  p2_init(&g11);
  p2_init(&g12);
  p6_init(&f0);
  p6_init(&f1);
  p6_init(&g0);
  p6_init(&g1);
  p6_set(&f0,f->p6coeff[0]);
  p6_set(&f1,f->p6coeff[1]);
  p2_set(&f00,(&f0)->p2coeff[0]);
  p2_set(&f01,(&f0)->p2coeff[1]);
  p2_set(&f02,(&f0)->p2coeff[2]);
  p2_set(&f10,(&f1)->p2coeff[0]);
  p2_set(&f11,(&f1)->p2coeff[1]);
  p2_set(&f12,(&f1)->p2coeff[2]);
  p6_set(&g0,g->p6coeff[0]);
  p6_set(&g1,g->p6coeff[1]);
  p2_set(&g00,(&g0)->p2coeff[0]);
  p2_set(&g01,(&g0)->p2coeff[1]);
  p2_set(&g02,(&g0)->p2coeff[2]);
  p2_set(&g10,(&g1)->p2coeff[0]);
  p2_set(&g11,(&g1)->p2coeff[1]);
  p2_set(&g12,(&g1)->p2coeff[2]);
  if((mpz_cmp((&g00)->coeff[0],(&f00)->coeff[0])==0)&&
     (mpz_cmp((&g00)->coeff[1],(&f00)->coeff[1])==0)&&
     (mpz_cmp((&g01)->coeff[0],(&f01)->coeff[0])==0)&&
     (mpz_cmp((&g01)->coeff[1],(&f01)->coeff[1])==0)&&
     (mpz_cmp((&g02)->coeff[0],(&f02)->coeff[0])==0)&&
     (mpz_cmp((&g02)->coeff[1],(&f02)->coeff[1])==0)&&
     (mpz_cmp((&g10)->coeff[0],(&f10)->coeff[0])==0)&&
     (mpz_cmp((&g10)->coeff[1],(&f10)->coeff[1])==0)&&
     (mpz_cmp((&g11)->coeff[0],(&f11)->coeff[0])==0)&&
     (mpz_cmp((&g11)->coeff[1],(&f11)->coeff[1])==0)&&
     (mpz_cmp((&g12)->coeff[0],(&f12)->coeff[0])==0)&&
     (mpz_cmp((&g12)->coeff[1],(&f12)->coeff[1])==0))
    res = 1;
  p2_clear(&f00);
  p2_clear(&f01);
  p2_clear(&f02);
  p2_clear(&f10);
  p2_clear(&f11);
  p2_clear(&f12);
  p2_clear(&g00);
  p2_clear(&g01);
  p2_clear(&g02);
  p2_clear(&g10);
  p2_clear(&g11);
  p2_clear(&g12);
  p6_clear(&f0);
  p6_clear(&f1);
  p6_clear(&g0);
  p6_clear(&g1);
  return res;
}


int main(int argc, char* argv[]){

  p2 Qx,Qy,Qz;
  p12 aopt,tmp,aoptZpr,aoptFr;
  mpz_t Fp,Fr,Zpr,Px,Py,Pz,Qx1,Qx2,Qy1,Qy2,Qz1,Qz2;
  ECPoint P;
  p2ECPoint Q,T;
  p12_init(&aopt);
  p12_init(&tmp);
  p12_init(&aoptZpr);
  p12_init(&aoptFr);
  EC_init_0(&P);
  p2EC_init(&Q);
  p2EC_init(&T);
  p2_init(&Qx);
  p2_init(&Qy);
  p2_init(&Qz);
  mpz_init(Px);
  mpz_init(Py);
  mpz_init(Pz);
  mpz_init(Qx1);
  mpz_init(Qy1);
  mpz_init(Qz1);
  mpz_init(Qx2);
  mpz_init(Qy2);
  mpz_init(Qz2);
  mpz_init(Fp);
  mpz_init(Fr);
  mpz_init(Zpr);

  mpz_set_str(Fp,"16798108731015832284940804142231733909889187121439069848933715426072753864723",10);
  mpz_set_str(Px,"2523648240000001ba344d80000000086121000000000013a700000000000012",16);
  mpz_set_str(Py,"1",10);
  mpz_set_str(Pz,"1",10);
  mpz_set_str(Fr,"4294967291",10);
  if(argc>1)
    mpz_set_str(Fr,argv[1],10);
  mpz_mul(Zpr,Fp,Fr);
  mpz_set_str(Qx1,"1c2141648fed8ba0f2a3febe8b98509bf86398d1fd1050c88fc3d88be3d15db1",16);
  mpz_set_str(Qx2,"93772aa06ea4acf488ce113f4a56aeb6c23264001c1501c1c59cd47faac6d0f",16);
  mpz_set_str(Qy1,"1eb672f0d5335990c9b12f9839b1a8804393211b198237c5acfc4d69d51186a0",16);
  mpz_set_str(Qy2,"118c5d037558e51efdd3cf3530d8c5cb65c52f9cf639ed6d81ddc6c16b76eec0",16);
  mpz_set_str(Qz1,"1",10);
  mpz_set_str(Qz2,"0",10);
  EC_set_mpz_t(&P,Px,Py,Pz);
  p2_set_mpz_t(&Qx,Qx1,Qx2);
  p2_set_mpz_t(&Qy,Qy1,Qy2);
  p2_set_mpz_t(&Qz,Qz1,Qz2);
  p2EC_set_p2(&Q,&Qx,&Qy,&Qz);
  
  OptAtePairingBN(&aoptZpr,&P,&Q,Fp,Zpr);
  OptAtePairingBN(&aoptFr,&P,&Q,Fp,Fr);
  
  p12in(&tmp,&aoptZpr,Fr);
  if(p12_is_eq(&tmp,&aoptFr))
    printf("Aucune faute détectée\n");
  else
    printf("Faute détectée\n");
  p12in(&aopt,&aoptZpr,Fp);
  printp12(&aopt, " <-- Résultat final\n",10);

  
  mpz_clear(Px);
  mpz_clear(Py);
  mpz_clear(Pz);
  mpz_clear(Qx1);
  mpz_clear(Qy1);
  mpz_clear(Qz1);
  mpz_clear(Qx2);
  mpz_clear(Qy2);
  mpz_clear(Qz2);
  p2_clear(&Qx);
  p2_clear(&Qy);
  p2_clear(&Qz);
  mpz_clear(Fp);
  mpz_clear(Fr);
  mpz_clear(Zpr);
  p12_clear(&aopt);
  p12_clear(&tmp);
  p12_clear(&aoptZpr);
  p12_clear(&aoptFr);
  EC_clear(&P);
  p2EC_clear(&Q);
  p2EC_clear(&T);
  return 0;
}
